unit MySQLClient;

{
  MySQL Native Delphi Client (c)2003 by Paul TOTH <tothpaul@free.fr>
  http://tothpaul.free.fr

  based on Mysql-direct library by Cristian Nicola

  v1.1 - bug fix for TMyField.Size

  v2.0 - add TMySQLSocket class
         add TMySQLFields class
         add TMySQLQuery  class
}

{

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

}

// http://public.logicacmg.com/~redferni/mysql/MySQL-Protocol.html

{$DEFINE COMPRESS} // use ZLIB (deflated) compression (inbound data only)
{$DEFINE RECVBUF } // use an internal receive cache buffer
{$DEFINE SENDBUF } // use an internal sending cache buffer

interface

uses
 Windows,Classes,SysUtils,Winsock,Math,Dialogs,zInflates;

type
 TMySQLFlag=(
  CLIENT_LONG_PASSWORD,
  CLIENT_FOUND_ROWS,
  CLIENT_LONG_FLAG,
  CLIENT_CONNECT_WITH_DB,
  CLIENT_NO_SCHEMA,
  CLIENT_COMPRESS,
  CLIENT_ODBC,
  CLIENT_LOCAL_FILES,
  CLIENT_IGNORE_SPACE,
  CLIENT_CHANGE_USER,
  CLIENT_INTERACTIVE,
  CLIENT_SSL,
  CLIENT_IGNORE_SIGPIPE,
  CLIENT_TRANSACTIONS
 );
 TMySQLFlags=set of TMySQLFlag;

 TMySQLCommand = (
  COM_SLEEP,        COM_QUIT,           COM_INIT_DB,
  COM_QUERY,        COM_FIELD_LIST,     COM_CREATE_DB,
  COM_DROP_DB,      COM_REFRESH,        COM_SHUTDOWN,
  COM_STATISTICS,   COM_PROCESS_INFO,   COM_CONNECT,
  COM_PROCESS_KILL, COM_DEBUG,          COM_PING,
  COM_TIME,         COM_DELAYED_INSERT, COM_CHANGE_USER,
  COM_BINLOG_DUMP,  COM_TABLE_DUMP,     COM_CONNECT_OUT
 );

 TMySQLFieldType = (
  dtDECIMAL,
  dtTINY,
  dtSHORT,
  dtLONG,
  dtFLOAT,
  dtDOUBLE,
  dtNULL,
  dtTIMESTAMP,
  dtLONGLONG,
  dtINT24,
  dtDATE,
  dtTIME,
  dtDATETIME,
  dtYEAR,
  dtNEWDATE,
  dtENUM=247,
  dtSET,
  dtTINY_BLOB,
  dtMEDIUM_BLOB,
  dtLONG_BLOB,
  dtBLOB,
  dtVAR_STRING,
  dtSTRING
 );

 TMySQLSocket=class
 private
  fHandle     :integer;
  fClientFlags:TMySQLFlags;
  fServerFlags:TMySQLFlags; // [CLIENT_LONG_FLAG..CLIENT_CONNECT_WITH_DB,CLIENT_COMPRESS]
  fBlockNumber:integer;
  fVersion    :byte;
  fServer     :string;
  fScramble   :string;
  fZCache :string;
 {$IFDEF RECVBUF}
  fRecvBuf:string;
  fRecvPos:integer;
  fRecvLen:integer;
 {$ENDIF}
 {$IFDEF SENDBUF}
  fSendBuf:string;
  fSendLen:integer;
  procedure SendBuffer;
 {$ENDIF}
  procedure Error(msg:string);
  procedure SetHandle(Value:integer);
  function inet(Host:pchar):integer;
  procedure SetFlag(Flag:TMySQLFlag; OnOff:boolean);
  procedure SetCheck(Value:integer);
  property Check:integer write SetCheck;
 protected
  function Crunch(Password:string):string;
  procedure ReadError;
  function BlockRead:string;
  procedure ZRead(var Data; Size:integer);
  procedure Read(var Data; Size:integer);
  procedure BlockWrite(var Data; Size:integer; Extra:integer=0);
  procedure Write(var Data; Size:integer);
  procedure Exec(Command:TMySQLCommand; Order:string);
  property Handle:integer read fHandle write SetHandle;
 public
  constructor Create; virtual;
  destructor Destroy; override;
  function Active:boolean;
  procedure Open(const AHostName:string; APort:integer=3306); virtual;
  procedure Close; virtual;
  procedure Login(const AUserName,APassword:string; ADataBase:string='');
  procedure Select(DataBase:string);
  function FoundRows:integer;
  procedure Ping;
 end;

 TMyField=class
 private
  fName  :string;
  fTable :string;
  fSize  :integer;
  fKind  :TMySQLFieldType;
  fFlags :word;
  fPrec  :byte;
  fValue :string;
  function GetAsInteger:integer;
 public
  property Name:string read fName write fName;
  property Table:string read fTable write fTable;
  property Size:integer read fSize write fSize;
  property Kind:TMySQLFieldType read fKind write fKind;
  property Flags:word read fFlags write fFlags;
  property Prec:byte read fPrec write fPrec;
  property AsString:string read fValue;
  property AsInteger:integer read GetAsInteger;
 end;

 TMySQLFields=class(TList)
 private
  function GetField(Index:integer):TMyField;
 public
  procedure Clear; override;
  function CreateFields(ASocket:TMySQLSocket):integer;
  procedure Bind(const Data:string);
  function FindField(const FieldName:string):TMyField;
  function FieldByName(const FieldName:string):TMyField;
  property Items[Index:integer]:TMyField read GetField; default;
 end;

 TMySQLQuery=class
 private
  fSocket:TMySQLSocket;
  fFields:TMySQLFields;
  fEof   :boolean;
 public
  constructor Create(ASocket:TMySQLSocket);
  destructor Destroy; override;
 // Direct Access
  function PrepareSQL(const SQL:string):integer;
  function FetchAll(Strings:TStrings):integer;
 // Fields
  function ExecSQL(const SQL:string):integer;
  procedure Fetch;
  procedure Bind(const Data:string);
  procedure Flush;
 // Utils
  function PrimaryKey(ATable:string):string;
  function FoundRows:integer;
 // properties
  property Fields:TMySQLFields read fFields;
  property Eof:boolean read fEof;
 end;

function SQLEncode(s:string):string;

implementation

resourcestring
 SSocketError='socket error';

//----------------------------------------------------------------------------//
procedure WSAStartup;
var
 wsa:TWSAData;
begin
 Winsock.WSAStartup($101,wsa);
end;

function SocketError:string;
var
 Value:integer;
begin
 Value:=WSAGetLastError;
 case Value of
  WSAECONNRESET : Result:='WSAECONNRESET : The virtual circuit was reset by the remote side';
  else Result:='WINSOCK Error #'+IntToStr(Value)+' (WSABASEERR+'+IntToStr(Value-WSABASEERR)+')';
 end;
end;

function SQLEncode(s:string):string;
var
 i,l,a:integer;
 c,d:char;
begin
 Result:=s;
 a:=0; l:=Length(s);
 for i:=1 to l do
  if s[i] in [#0,#10,#13,#26,'''','"','\'] then begin
   Result[i]:='\';
   inc(a);
  end;
 if a=0 then exit; // no escape needed
 i:=l+a;
 SetLength(Result,i); // reserve space for "\"
 FillChar(Result[l+1],a,' '); // initialiser les nouveaux caractres !
 while a>0 do begin
  c:=Result[l];
  if c='\' then begin // set by the above loop
   d:=s[l]; // original char
   case d of
    #00: d:='0';
    #10: d:='n';
    #13: d:='r';
    #26: d:='Z';
   end;
   Result[i]:=d; dec(i);
   dec(a);
  end;
  Result[i]:=c; dec(i);
  dec(l);
 end;
end;

procedure hash(s:string; var res0,res1:longint);
var
 nr1,add,nr2,tmp:int64;
 i:longint;
 e1:int64;
// len:longint;
begin
 nr1:=1345345333;
 add:=7;
 nr2:=$12345671;
 for i:=1 to length(s) do begin
  if (s[i]=#20)or(s[i]=#9) then continue;
  tmp:=ord(s[i]);
  e1 :=(((nr1 and 63)+add)*tmp)+(nr1 shl 8);
  nr1:=nr1 xor e1;
 {$IFOPT Q+}{$DEFINE KEEPQ}{$ENDIF}{$Q-} // thanks Jiri Barton
  nr2:=nr2+((nr2 shl 8) xor nr1);
 {$IFDEF KEEPQ}{$Q+}{$UNDEF KEEPQ}{$ENDIF} // thanks Jiri Barton
  inc(add,tmp);
 end;
 res0:=nr1 and $7fffffff;
 res1:=nr2 and $7fffffff;
end;

function VarSize(const Data:string; var Index:integer):integer;
var
 Code:integer;
begin
 if Length(Data)<Index then
  Result:=-1
 else begin
  Code:=ord(Data[Index]); inc(Index);
  case Code of
   251 : begin Result:=-1; exit; end;
   252 : begin Result:=pword(@Data[Index])^; inc(Index,2); end;
   253 : begin Result:=pinteger(@Data[Index])^ and $FFFFFF; inc(Index,3); end;
   254 : begin
          if (pinteger(@Data[Index])^)<>0 then raise Exception.Create('BIG DATA not supported');
          Result:=pinteger(@Data[Index+4])^;
          inc(Index,8);
         end;
   else  begin Result:=Code; exit; end;
  end;
 end;
end;

function VarData(const Data:string; var Index:integer):string;
var
 Len:integer;
begin
 Len:=VarSize(Data,Index);
 if Len<0 then
  Result:=''
 else begin
  Result:=Copy(Data,Index,Len);
  inc(Index,Len);
 end;
end;

//----------------------------------------------------------------------------//
{ TMyField }
//----------------------------------------------------------------------------//

function TMyField.GetAsInteger:integer;
begin
 if fValue='' then Result:=0 else Result:=StrToInt(fValue);
end;

{------------------------------------------------------------------------------}
{ TMySQLSocket }
{------------------------------------------------------------------------------}

constructor TMySQLSocket.Create;
begin
 inherited;
 fHandle:=INVALID_SOCKET;
 fClientFlags:=[CLIENT_LONG_PASSWORD,CLIENT_LONG_FLAG,CLIENT_LOCAL_FILES,CLIENT_TRANSACTIONS];
end;

destructor TMySQLSocket.Destroy;
begin
 close;
 inherited;
end;

function TMySQLSocket.inet(Host: pchar): integer;
var
 PHost:PHostEnt;
begin
 Result:=inet_addr(Host);
 if Result=INADDR_NONE then begin
  PHost:=gethostbyname(Host);
  if PHost=nil then Error('Host not found');
  Result:=integer(pointer(PHost.h_addr^)^);
 end;
end;

procedure TMySQLSocket.Error(msg: string);
begin
 raise Exception.Create(msg);
end;

procedure TMySQLSocket.SetCheck(Value: integer);
begin
 if Value=SOCKET_ERROR then Error(SocketError);
end;

procedure TMySQLSocket.SetHandle(Value: integer);
begin
 if Value=INVALID_SOCKET then Error(SSocketError);
 fHandle:=Value;
end;

procedure TMySQLSocket.Open(const AHostName: string; APort: integer);
var
 Sock:TSockAddr;
 Data:string;
 i,l:integer;
begin
 Close;
 Handle:=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
 try
 {$IFDEF RECVBUF}
  l:=SizeOf(i);
  if getsockopt(fHandle,SOL_SOCKET,SO_RCVBUF,pchar(@i),l)=SOCKET_ERROR then i:=1024;
  SetLength(fRecvBuf,i);
  fRecvLen:=0;
 {$ENDIF}
 {$IFDEF SENDBUF}
  l:=SizeOf(i);
  if getsockopt(fHandle,SOL_SOCKET,SO_SNDBUF,pchar(@i),l)=SOCKET_ERROR then
   i:=512
  else
   if i>512 then i:=512;
  SetLength(fSendBuf,i);
  fSendLen:=0;
 {$ENDIF}
  FillChar(Sock,SizeOf(Sock),0);
  Sock.sin_family:=AF_INET;
  Sock.sin_port:=htons(APort);
  Sock.sin_addr.S_addr:=inet(pchar(AHostName));
  Check:=connect(fHandle,Sock,SizeOf(Sock));
  Data:=BlockRead;
  fVersion :=ord(Data[1]);
  if fVersion<>10 then Error('Invalide version');
  fServer:=pchar(@Data[2]);
  i:=1+Length(fServer)+1+5;
  fScramble:=Copy(Data,i,8);
  Move(Data[i+9],fServerFlags,SizeOf(fServerFlags));
 except
  Close;
 end;
end;

procedure TMySQLSocket.Close;
begin
 if fHandle<>INVALID_SOCKET then begin
  closesocket(fHandle);
  fHandle:=INVALID_SOCKET;
 end;
end;

function TMySQLSocket.Active:boolean;
begin
 Result:=fHandle<>INVALID_SOCKET;
end;

procedure TMySQLSocket.Ping;
begin
 Exec(COM_PING,'');
 ReadError;
end;

function TMySQLSocket.Crunch(Password: string): string;
var
 hp0,hp1:longint;
 hm0,hm1:longint;
 maxValue,seed, seed2 :int64;
 dRes: double;
 i:longint;
 e:byte;
len1:longint;
begin
 if (Password='') then
  Result:=''
 else begin
  len1:=Length(fScramble);
  SetLength(Result,len1);
  hash(Password ,hp0,hp1);
  hash(fScramble,hm0,hm1);
  maxValue:= $3FFFFFFF;
  seed  := ( hp0 xor hm0 )mod maxValue ;
  seed2 := ( hp1 xor hm1 )mod maxValue ;
  for i:=1 to len1 do begin
   seed  := ( seed * 3 + seed2  )mod maxValue ;
   seed2 := ( seed + seed2 + 33 )mod maxValue ;
   dRes := Seed / maxValue;
   result[i] := char( Floor( dRes * 31 ) + 64 );
  end;
  seed:=(seed * 3 + seed2) mod maxValue ;
  dRes:= Seed / maxValue;
  e := floor( dRes * 31 );
  for i := 1 to len1 do result[i]:=chr( byte(result[i]) xor e);
 end;
end;

procedure TMySQLSocket.SetFlag(Flag:TMySQLFlag; OnOff:boolean);
begin
 if OnOff then Include(fClientFlags,Flag) else Exclude(fClientFlags,Flag);
end;

procedure TMySQLSocket.Login(const AUserName, APassword: string;
  ADataBase: string);
var
 Data:string;
 Flags:TMySQLFlags;
begin
// send : Flags (2 bytes), MAX_ALLOWED_PACKET (3 bytes), UserZ, PassZ, DataBaseZ
 Data:=#0#0+#0#0#0+AUserName+#0+Crunch(APassword)+#0+ADataBase+#0;
 SetFlag(CLIENT_CONNECT_WITH_DB,ADataBase<>'');
 Flags:=fClientFlags;
{$IFDEF COMPRESS}
// do not set CLIENT_COMPRESS FLAG !!! Yet !!!
 if CLIENT_COMPRESS in fServerFlags then Include(Flags,CLIENT_COMPRESS);
{$ENDIF}
 Move(Flags,Data[1],2);
 BlockWrite(Data[1],Length(Data));
 ReadError;
 fClientFlags:=Flags; { Now we can set CLIENT_COMPRESS (see BlockRead) }
 if not (CLIENT_CONNECT_WITH_DB in fServerFlags) then Select(ADataBase);
end;

procedure TMySQLSocket.Select(DataBase:string);
begin
 Exec(COM_INIT_DB,DataBase);
 ReadError;
end;

function TMySQLSocket.BlockRead: string;
var
 Size:integer; // 3 bytes + BlockNumber
begin
{$IFDEF SENDBUF}
 SendBuffer; // make sure that the request is send before ;D
{$ENDIF}
 if CLIENT_COMPRESS in fClientFlags then begin
  ZRead(Size,SizeOf(Size));
  SetLength(Result,Size);
  ZRead(Result[1],Size);
 end else begin
  Read(Size,SizeOf(Size));
  if Size shr 24<>fBlockNumber then Error('Protocol error #1');
  fBlockNumber:=Succ(fBlockNumber) and $FF;
  Size:=Size and $FFFFFF;
  SetLength(Result,Size);
  Read(Result[1],Size);
 end;
 if Result[1]=#255 then Error('MySQL Error #'+IntToStr(pword(@Result[2])^)+' : '+Copy(Result,4,Size));
end;

procedure TMySQLSocket.BlockWrite(var Data; Size, Extra: integer);
var
 Head:integer;
 Comp:record
  RealSize:integer;
  CompSize:integer;
 end;
begin
 if CLIENT_COMPRESS in fClientFlags then begin
  Comp.RealSize:=(fBlockNumber shl 24) or (SizeOf(Head)+Size+Extra);
  fBlockNumber:=Succ(fBlockNumber) and $FF;
  Comp.CompSize:=0; // do not compress...
  Write(Comp,7);    // CompSize = 3 bytes only
  Head:=Size+Extra; // BlockNumber = 0
 end else begin
  Head:=(fBlockNumber shl 24) or (Size+Extra);
  fBlockNumber:=Succ(fBlockNumber) and $FF;
 end;
 Write(Head,SizeOf(Head));
 Write(Data,Size);
end;

procedure TMySQLSocket.ReadError;
var
 Data:string;
begin
 Data:=BlockRead;
 if copy(Data,1,3)<>#0#0#0 then Error('Unknow error from server');
end;

procedure TMySQLSocket.Exec(Command: TMySQLCommand; Order: string);
var
 Extra:integer;
begin
 fBlockNumber:=0;
 Extra:=Length(Order);
 BlockWrite(Command,1,Extra);
 Write(Order[1],Extra);
end;

procedure TMySQLSocket.Read(var Data; Size: integer);
var
 p:pchar;
 i:integer;
begin
 p:=@Data;
 while Size>0 do begin
{$IFDEF RECVBUF}
  if fRecvLen=0 then begin
   fRecvLen:=recv(fHandle,fRecvBuf[1],Length(fRecvBuf),0);
   if fRecvLen<=0 then begin
    Close;
    Error('Read error '+SocketError);
   end;
   fRecvPos:=0;
  end;
  i:=fRecvLen-fRecvPos;
  if i>Size then i:=Size;
  Move(fRecvBuf[fRecvPos+1],p[0],i);
  inc(fRecvPos,i);
  if fRecvPos=fRecvLen then fRecvLen:=0;
{$ELSE}
  i:=recv(fHandle,p[0],Size,0);
  if i<=0 then Error('Read error '+SocketError);
{$ENDIF}
  inc(p,i);
  dec(Size,i);
 end;
end;

{$IFDEF SENDBUF}
procedure TMySQLSocket.SendBuffer;
var
 p:pchar;
 i:integer;
begin
 p:=@fSendBuf[1];
 while fSendLen>0 do begin
  i:=send(fHandle,p[0],fSendLen,0);
  if i<=0 then begin
   Close;
   Error('Write error '+SocketError);
  end;
  dec(fSendLen,i);
  inc(p,i);
 end;
end;
{$ENDIF}

procedure TMySQLSocket.Write(var Data; Size: integer);
var
 p:pchar;
 i:integer;
begin
 p:=@Data;
 while Size>0 do begin
 {$IFDEF SENDBUF}
  i:=Length(fSendBuf)-fSendLen;
  if i>Size then i:=Size;
  move(p[0],fSendBuf[fSendLen+1],i);
  inc(fSendLen,i);
  if fSendLen=Length(fSendBuf) then SendBuffer;
 {$ELSE}
  i:=send(fHandle,p[0],Size,0);
  if i<=0 then begin
   Close;
   Error('Write error '+SocketError);
  end;
 {$ENDIF}
  inc(p,i);
  dec(Size,i);
 end;
end;

procedure TMySQLSocket.ZRead(var Data; Size: integer);
var
 Len :integer;
 Head:record
  CompSize:integer; // 3 bytes + BlockNumber
  RealSize:integer; // 3 bytes only !!
 end;
 Comp:string;
begin
 Len:=Length(fZCache);
 while Len<Size do begin
  Head.RealSize:=0; // 4th byte
  Read(Head,7);     // compression Header
  if Head.CompSize shr 24<>fBlockNumber then Error('ZProtocol error #1');
  fBlockNumber:=Succ(fBlockNumber) and $FF;
  Head.CompSize:=Head.CompSize and $FFFFFF;
  if Head.RealSize=0 then begin // no compression
   SetLength(fZCache,Len+Head.CompSize);
   Read(fZCache[Len+1],Head.CompSize);
   inc(Len,Head.CompSize);
  end else begin
   SetLength(Comp,Head.CompSize);
   Read(Comp[1],Head.CompSize);
   SetLength(fZCache,Len+Head.RealSize);
  // Check DEFLATED compression...hope it's ok :}
   if (Comp[1]<>#$78)or(Comp[2]<>#$9C) then Error('ZProtocol error #2');
   Inflate(Comp[3],fZCache[Len+1]);
   inc(Len,Head.RealSize);
  end;
 end;
 Move(fZCache[1],Data,Size);
 if Size=Len then fZCache:='' else Delete(fZCache,1,Size);
end;

function TMySQLSocket.FoundRows: integer;
var
 s:string;
 x:integer;
begin
 Exec(COM_QUERY,'SELECT FOUND_ROWS()');
 x:=1;
 s:=BlockRead;
 if VarSize(s,x)<>1 then Error('FOUND_ROWS() error !');
 s:=BlockRead; // Field Desc
 if BlockRead<>#254 then Error('FoundRows error #1');
 x:=1;
 s:=BlockRead;
 Result:=StrToInt(VarData(s,x));
 if BlockRead<>#254 then Error('FoundRows error #2');
end;

{ TMySQLFields }

procedure TMySQLFields.Bind(const Data: string);
var
  i,x:integer;
begin
 x:=1;
 for i:=0 to Count-1 do Items[i].fValue:=VarData(Data,x);
end;

procedure TMySQLFields.Clear;
var
 i:integer;
begin
 for i:=0 to Count-1 do Items[i].Free;
 inherited;
end;

function TMySQLFields.CreateFields(ASocket: TMySQLSocket):integer;
var
 s:string;
 x:integer;
 f:TMyField;
begin
 Result:=0;
 s:=ASocket.BlockRead;
 while s<>#254 do begin
  inc(Result);
  x:=1;
  f:=TMyField.Create; Add(f);
  f.fTable:=VarData(s,x);
  f.fName :=VarData(s,x);
  inc(x);   // #3
  f.fSize:=PInteger(@s[x])^ and $FFFFFF;
  inc(x,4); // Size, #1
  f.fKind:=TMySQLFieldType(s[x]);
  inc(x);
  if s[x]=#2 then begin
   f.fFlags:=ord(s[x+1]);
   inc(x,2);
  end else begin
   f.fFlags:=PWord(@s[x+1])^;
   inc(x,3);
  end;
  f.fPrec:=ord(s[x]);
  if x<>Length(s) then raise Exception.Create('Error while reading field def');
  s:=ASocket.BlockRead;
 end;
end;

function TMySQLFields.FieldByName(const FieldName: string): TMyField;
begin
 Result:=FindField(FieldName);
 if Result=nil then raise Exception.Create('Field "'+FieldName+'" not found');
end;

function TMySQLFields.FindField(const FieldName: string): TMyField;
var
 i:integer;
begin
 i:=Count-1;
 while (i>=0) do begin
  Result:=Items[i];
  if AnsiCompareText(Result.Name,FieldName)=0 then exit;
  dec(i);
 end;
 Result:=nil;
end;

function TMySQLFields.GetField(Index: integer): TMyField;
begin
 Result:=Get(Index);
end;

//----------------------------------------------------------------------------//

{ TMySQLQuery }

constructor TMySQLQuery.Create(ASocket:TMySQLSocket);
begin
 fSocket:=ASocket;
 fFields:=TMySQLFields.Create;
 fEof:=True;
end;

destructor TMySQLQuery.Destroy;
begin
 Flush;
 fFields.Free;
 inherited;
end;

function TMySQLQuery.PrepareSQL(const SQL: string): integer;
var
 s:string;
 x:integer;
begin
 Flush;
 fSocket.Exec(COM_QUERY,SQL);
 s:=fSocket.BlockRead;
 x:=1;
 Result:=VarSize(s,x); // Field count
 if Result=0 then begin
 // INSERT, UPDATE...no data returned
  VarSize(s,x); // Affected Rows
  // InsertID (VarSize)
  Result:=VarSize(s,x);
  // Status [if CLIENT_TRANSACTIONS in fServerFlags] (Word)
  // Message (AsciiZ)
 end else begin
 // SELECT with returned data
  if fFields.CreateFields(fSocket)<>Result then fSocket.Error('PrepareSQL error');
  fEof:=False;
 end;
end;

function TMySQLQuery.FetchAll(Strings: TStrings):integer;
var
 s:string;
begin
 Result:=Strings.Count;
 while not fEof do begin
  s:=fSocket.BlockRead;
  if s=#254 then fEof:=True else Strings.add(s);
 end;
 Result:=Strings.Count-Result;
end;

function TMySQLQuery.ExecSQL(const SQL: string): integer;
begin
 Result:=PrepareSQL(SQL);
 Fetch;
end;

procedure TMySQLQuery.Fetch;
begin
 if not fEof then Bind(fSocket.BlockRead);
end;

procedure TMySQLQuery.Bind(const Data:string);
begin
 if Data=#254 then fEof:=True else fFields.Bind(Data);
end;

procedure TMySQLQuery.Flush;
begin
 while not fEof do fEof:=(fSocket.BlockRead=#254);
 fFields.Clear;
end;

function TMySQLQuery.PrimaryKey(ATable:string):string;
begin
 Result:='';
 ExecSQL('DESC '+ATable);
 while not fEof do begin
  if Fields[3].AsString='PRI' then begin
   if Result<>'' then Result:=Result+', ';
   Result:=Result+Fields[0].AsString;
  end;
  Fetch;
 end;
 Flush;
end;

function TMySQLQuery.FoundRows: integer;
begin
 if not fEof then fSocket.Error('Fetch all data before');
 Result:=fSocket.FoundRows;
end;

initialization
 WSAStartup;
finalization
 WSACleanup;
end.
