unit MySqlDS;

{
  MySQL Native Delphi DataSet (c)2003 by Paul TOTH <tothpaul@free.fr>
  http://tothpaul.free.fr
}

{

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

}

// http://www.btinternet.com/~joannac/keephold.htm

interface

uses
 Windows,Classes,SysUtils,Forms,DB,MySQLClient;

type
 TMySQLRecord=record
  Index:integer;       // nous ne conservons que l'index dans la table, pas les donnes
  Flag :TBookmarkFlag; // le Flag pour Delphi
 end;

 TMyTable=class(TMySQLQuery)
 private
  fTable  :string;         // nom de la table
  fOrder  :string;         // ordre de trie
  fCache  :TStringList;    // enregistrements en cache
  fPage   :integer;        // nombre d'lments lus (LIMIT)
  fFirst  :integer;        // n du premier enregistrement (LIMIT)
  fLast   :integer;        // n du dernier enregistrement (LIMIT)
  fCount  :integer;        // nombre total d'enregistrements (FOUND_ROWS)
  fCursor :integer;        // Numro de l'enregistrement en cours
  procedure NextPage;
  procedure PriorPage;
  function LoadPriorPage:integer;
  procedure SetCursor(Value:integer);
 public
  constructor Create(ASocket:TMySQLSocket);
  destructor Destroy; override;
  procedure SetTable(ATable,KeyFields:string);
  procedure First;
  procedure Last;
  procedure Refresh(Current:integer);
  function Next:TGetResult;
  function Prior:TGetResult;
  property Cursor:integer read fCursor write SetCursor;
 end;

 TMySQLConnection=class(TCustomConnection)
 private
  fHostName :string; // serveur MySQL
  fUserName :string; // utilisateur
  fPassword :string; // mot de passe
  fDatabase :string; // nom de la base MySQL
  fSocket   :TMySQLSocket;
  procedure SetParam(Index:integer;const Value:string);
  procedure SetDatabase(Value:string);
 protected
  procedure DoConnect; override;
  procedure DoDisconnect; override;
  function GetConnected: Boolean; override;
 public
  destructor Destroy; override;
  procedure ShowQuery(const Query:string; List:TStrings);
 published
  property HostName:string index 0 read fHostName write SetParam;
  property UserName:string index 1 read fUserName write SetParam;
  property Password:string index 2 read fPassword write SetParam;
  property Database:string read fDatabase write SetDatabase;
  property Connected;
 end;

 TMySQLDataSet=class(TDataSet)
 private
  fConnection:TMySQLConnection;
  fTableName :string;
  fMyTable   :TMyTable;
  fCurrent   :integer;
  fBufSize   :integer;      // taille du buffer de donnes
  fBookMarkSize:integer;    // taille du Bookmark
  procedure SetConnection(Value:TMySQLConnection);
  procedure SetTableName(Value:string);
  procedure ConnectChange(Sender: TObject; Connecting: Boolean);
 protected
  fIsOpen:boolean;
  function  GetRecordCount: Integer; override;
  function  GetRecNo:integer; override;
  procedure SetRecNo(Value:integer); override;
 { abstract methods required for all datasets }
  procedure InternalOpen; override;
  procedure InternalClose; override;
  function IsCursorOpen: Boolean; override;
  procedure InternalInitFieldDefs; override;
  function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
  procedure InternalHandleException; override;
 { Method stubs for UniDirectional/Readonly/Unbuffered datasets }
  function AllocRecordBuffer: PChar; override;
  procedure FreeRecordBuffer(var Buffer: PChar); override;
  procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
  function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
  function GetRecordSize: Word; override;
  procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
  procedure InternalDelete; override;
  procedure InternalFirst; override;
  procedure InternalRefresh; override;
  procedure InternalGotoBookmark(Bookmark: Pointer); override;
  procedure InternalInitRecord(Buffer: PChar); override;
  procedure InternalLast; override;
  procedure InternalPost; override;
  procedure InternalSetToRecord(Buffer: PChar); override;
  procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
  procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
  procedure SetFieldData(Field: TField; Buffer: Pointer); override;
 public
  constructor Create(AOwner:TComponent); override;
  function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
 published
  property Connection:TMySQLConnection read fConnection write SetConnection;
  property TableName:string read fTableName write SetTableName;
  property Active;
 end;

implementation

{ TMyTable }

constructor TMyTable.Create(ASocket:TMySQLSocket);
begin
 inherited;
 fCache:=TStringList.Create;
 fPage:=10;
end;

destructor TMyTable.Destroy;
begin
 fCache.Free;
 inherited;
end;

procedure TMyTable.SetTable(ATable, KeyFields: string);
begin
 fTable:=ATable;
 fOrder:=KeyFields;
 if fOrder='' then begin  // cl primaire
  fOrder:=PrimaryKey(fTable);
  if fOrder='' then raise Exception.Create('pas de cl primaire');
 end;
 First;
end;

procedure TMyTable.First;
begin
 fCache.Clear;
 fFirst:=0;
 fLast:=-1;
 NextPage;
 fCursor:=-1;
end;

procedure TMyTable.Last;
var
 i:integer;
begin
 fCache.Clear;
 PrepareSQL('SELECT SQL_CALC_FOUND_ROWS * FROM '+fTable+' ORDER BY '+fOrder+' DESC LIMIT '+IntToStr(fPage));
 i:=LoadPriorPage;
 fCount:=FoundRows;
 fLast :=fCount-1;
 fFirst:=fCount-i;
 fCursor:=-1;
end;

function TMyTable.Next: TGetResult;
begin
 if fCursor=fLast-1 then NextPage;
 if fCursor<=fLast-1 then begin
  inc(fCursor);
  Result:=grOK;
 end else begin
  Result:=grEOF;
 end;
end;

function TMyTable.Prior: TGetResult;
begin
 if (fFirst>0)and(fCursor=fFirst) then PriorPage;
 if fCursor>0 then begin
  dec(fCursor);
  Result:=grOK;
 end else begin
  Result:=grBOF;
 end;
end;

procedure TMyTable.NextPage;
begin
 PrepareSQL('SELECT SQL_CALC_FOUND_ROWS * FROM '+fTable+' ORDER BY '+fOrder+' LIMIT '+IntToStr(fLast+1)+','+IntToStr(fPage));
 inc(fLast,FetchAll(fCache));
 fCount:=FoundRows;
end;

procedure TMyTable.PriorPage;
begin
 PrepareSQL('SELECT * FROM '+fTable+' ORDER BY '+fOrder+' DESC LIMIT '+IntToStr(fCount-fFirst)+','+IntToStr(fPage));
 dec(fFirst,LoadPriorPage);
 if fFirst<0 then begin
  dec(fCount,fFirst);
  dec(fLast,fFirst);
  fFirst:=0;
 end;
end;

procedure TMyTable.SetCursor(Value:integer);
begin
 fCursor:=Value;
 if (fCursor<fFirst)or(fCursor>fLast) then begin
  fCache.Clear;
  fFirst:=fCursor-fPage div 2;
  if fFirst<0 then fFirst:=0;
  fLast :=fFirst-1;
  NextPage;
 end;
 Fields.Bind(fCache[fCursor-fFirst]);
end;

function TMyTable.LoadPriorPage:integer;
var
 l:TStringList;
 i:integer;
begin
 l:=TStringList.Create;
 try
  l.Capacity:=fPage;
  Result:=FetchAll(l);
  fCache.Capacity:=fCache.Count+Result;
  for i:=0 to Result-1 do fCache.Insert(0,l[i]);
 finally
  l.Free;
 end;
end;

procedure TMyTable.Refresh(Current:integer);
begin
 fCache.Clear;
 fCursor:=-1;
 fFirst:=Current-fPage div 2;
 if fFirst<0 then fFirst:=0;
 fLast :=fFirst-1;
 NextPage;
end;

{ TMySQLConnexion }

destructor TMySQLConnection.Destroy;
begin
 inherited;
 fSocket.Free;
end;

procedure TMySQLConnection.DoConnect;
begin
 if fHostName='' then raise Exception.Create('HostName not set');
 fSocket:=TMySQLSocket.Create;
 fSocket.Open(fHostName);
 fSocket.Login(fUserName,fPassword,fDatabase);
end;

procedure TMySQLConnection.DoDisconnect;
begin
 FreeAndNil(fSocket);
end;

function TMySQLConnection.GetConnected: Boolean;
begin
 Result:=(fSocket<>nil);
end;

procedure TMySQLConnection.SetDatabase(Value: string);
begin
 fDatabase:=Value;
 if Connected then fSocket.Select(fDatabase);
end;

procedure TMySQLConnection.SetParam(Index: integer;const Value: string);
begin
 Close;
 case Index of
  0: fHostName:=Value;
  1: fUserName:=Value;
  2: fPassword:=Value;
 end;
end;

procedure TMySQLConnection.ShowQuery(const Query:string; List: TStrings);
var
 q:TMySQLQuery;
begin
 if not Connected then exit;
 q:=TMySQLQuery.Create(fSocket);
 try
  q.ExecSQL(Query);
  while not q.Eof do begin
   List.Add(q.Fields[0].AsString);
   q.Fetch;
  end;
 finally
  q.Free;
 end;
end;

{ TMySQLDataSet }

constructor TMySQLDataSet.Create(AOwner: TComponent);
begin
 inherited;
 fCurrent:=-1;
end;

procedure TMySQLDataSet.SetTableName(Value: string);
begin
 Close;
 fTableName:=Value;
end;

procedure TMySQLDataSet.InternalOpen;
begin
 if fConnection=nil then raise Exception.Create('Connection not set');
 fConnection.Open;
 fMyTable:=TMyTable.Create(fConnection.fSocket);
 try
  fMyTable.SetTable(fTableName,'');
  fCurrent:=-1;
  InternalInitFieldDefs;
  if DefaultFields then CreateFields;
  BindFields(True);
  fBufSize:=SizeOf(TMySQLRecord)+CalcFieldsSize;
  fBookmarkSize:=SizeOf(Integer);
 except
  FreeAndNil(fMyTable);
 end;
end;

procedure TMySQLDataSet.InternalClose;
begin
 BindFields(False);
 if DefaultFields then DestroyFields;
 FreeAndNil(fMyTable);
 fCurrent:=-1;
end;

function TMySQLDataSet.IsCursorOpen: Boolean;
begin
 Result:=fMyTable<>nil;
end;

procedure TMySQLDataSet.InternalInitFieldDefs;
var
 i:integer;
begin
 FieldDefs.Clear;
 for i:=0 to fMyTable.Fields.Count-1 do begin
  with fMyTable.Fields[i] do FieldDefs.Add(Name,ftString,Size);
 end;
end;

function TMySQLDataSet.GetRecord(Buffer: PChar; GetMode: TGetMode;
  DoCheck: Boolean): TGetResult;
begin
 Result:=grError;
 if fMyTable.fCount=0 then exit;
 case GetMode of
  gmCurrent : if fCurrent<0 then Result:=grEOF else begin
               fMyTable.Cursor:=fCurrent;
               Result:=grOK;
              end;
  gmNext    : begin
   fMyTable.fCursor:=fCurrent;
   Result:=fMyTable.Next;
  end;
  gmPrior   : begin
   fMyTable.fCursor:=fCurrent;
   Result:=fMyTable.Prior;
  end;
  else Result:=grError;
 end;
 if Result=grOK then begin
  with TMySQLRecord(pointer(Buffer)^) do begin
   fCurrent:=fMyTable.fCursor;
   Index:=fMyTable.fCursor;
   Flag :=bfCurrent;
  end;
  GetCalcFields(Buffer); // champs calcules, c'est Delphi qui bosse
 end;
end;

procedure TMySQLDataSet.InternalHandleException;
begin
 Application.HandleException(Self);
end;

function TMySQLDataSet.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
 i:integer;
 s:string;
begin
 if (Field.FieldNo=0)or(Field.FieldNo>fMyTable.Fields.Count) then begin
  Result:=False;
  exit;
 end;
 fMyTable.Cursor:=TMySQLRecord(pointer(ActiveBuffer)^).Index;
 with fMyTable.Fields[Field.FieldNo-1] do begin
  s:=AsString;
  i:=Length(s);
  if i>=Field.DataSize then i:=Field.DataSize-1;
  Move(s[1],Buffer^,i);
  PChar(Buffer)[i]:=#0;
 end;
 Result:=True;
end;

function TMySQLDataSet.GetRecordCount: Integer;
begin
 Result:=fMyTable.fCount;
end;

function TMySQLDataSet.GetRecNo: integer;
begin
 UpdateCursorPos; // s'assure qu'on est sur le record courrant
 Result:=fCurrent+1;
end;

procedure TMySQLDataSet.SetRecNo(Value: integer);
begin
 CheckBrowseMode; // s'assure que la source de donnes n'est pas en dition...
 fCurrent:=Value-1;
 Resync([rmCenter]); // repositionner (la DBGrid)
end;

function TMySQLDataSet.AllocRecordBuffer: PChar;
begin
 Result:=AllocMem(fBufSize);
end;

procedure TMySQLDataSet.FreeRecordBuffer(var Buffer: PChar);
begin
 FreeMem(Buffer);
end;

procedure TMySQLDataSet.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
 Move(Buffer^,Data^,fBookmarkSize);
end;

function TMySQLDataSet.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
 Result:=TMySqlRecord(pointer(Buffer)^).Flag;
end;

function TMySQLDataSet.GetRecordSize: Word;
begin
 Result:=0;
end;

procedure TMySQLDataSet.InternalAddRecord(Buffer: Pointer; Append: Boolean);
begin
 inherited;
end;

procedure TMySQLDataSet.InternalDelete;
begin
 inherited;
end;

procedure TMySQLDataSet.InternalFirst;
begin
 fMyTable.First;
 fCurrent:=-1;
end;

procedure TMySQLDataSet.InternalGotoBookmark(Bookmark: Pointer);
begin
 fCurrent:=integer(Bookmark^);
end;

procedure TMySQLDataSet.InternalInitRecord(Buffer: PChar);
begin
 FillChar(Buffer^,fBufSize,0);
 inherited;
end;

procedure TMySQLDataSet.InternalLast;
begin
 fMyTable.Last;
 fCurrent:=fMyTable.fCount{-1};
end;

procedure TMySQLDataSet.InternalPost;
begin
 inherited;
end;

procedure TMySQLDataSet.InternalSetToRecord(Buffer: PChar);
begin
 fCurrent:=TMySQLRecord(pointer(Buffer)^).Index;
end;

procedure TMySQLDataSet.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
 Move(Data^,Buffer^,fBookMarkSize);
end;

procedure TMySQLDataSet.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
 TMySQLRecord(pointer(Buffer)^).Flag:=Value;
end;

procedure TMySQLDataSet.SetFieldData(Field: TField; Buffer: Pointer);
begin
 inherited;
end;

procedure TMySQLDataSet.InternalRefresh;
begin
 fMyTable.Refresh(fCurrent);
end;

procedure TMySQLDataSet.SetConnection(Value: TMySQLConnection);
begin
 if fConnection<>nil then begin
  Close;
  fConnection.UnRegisterClient(Self);
 end;
 fConnection:=Value;
 if fConnection<>nil then begin
  fConnection.RegisterClient(Self,ConnectChange);
 end;
end;

procedure TMySQLDataSet.ConnectChange(Sender: TObject; Connecting: Boolean);
begin
 //if not Connecting then Close;
end;

end.
